/****************************************************************************
 File name:  CS460_PthreadsLab.c
 Author:     chadd williams
 Date:       March 7, 2012
 Class:      CS460
 Assignment: PThreads Lab
 Purpose:	   Practice synchronization
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <unistd.h>

#include <stdbool.h>

// global shared data
long double gSum = 0.0;
int *gpaThreadExecutionCounter;

// tunable parameters
bool gbSynchronizationOn;
int gMaxTids;
int gWorkMultiplier;


/****************************************************************************
Function: 		runner

Description: 	The function run during thread invocation.
							update the global value by 0.1 some number of times.

Parameters: 	pParam -  a pointer to an integer representing the user generated
							Thread ID.

Returned: 		NULL
****************************************************************************/
void* runner(void* pParam)
{
	int tid = *(int*) pParam;
	int i;

	// tids start at zero!
	for(i=0 ; i < gWorkMultiplier * (tid+1) ; i++) 
	{

		//update the global counter
		gSum += .1;
		
		//fprintf(stderr,"[%d] adding to the global! %Lf\n",tid, gSum);
		//printf("[%d] adding to the global! %Lf\n",tid, gSum);

		//track how many times each thread updates the global variable
		gpaThreadExecutionCounter[tid] ++;
	}

	return NULL;
}


/****************************************************************************
Function: 		parseCommandLineOptions

Description: 	Parse the command line options and fill the parameters.

Parameters: 	argc, argv: take command line options
							pbSynchronization: -x
							pMaxTids: number of threads
							pWorkMultipler: work multiplier option
							
Returned: 		NULL
****************************************************************************/

void parseCommandLineOptions(int argc, char** argv, bool *pbSynchronization, 
	int *pMaxTids, int *pWorkMultiplier)
{
	
	int option;
	char* pOptValue;
	
	while( -1 != (option = getopt(argc, argv, "xm:w:")))
	{
		switch(option)
		{
			case 'x':
				*pbSynchronization = false;
				break;
			case 'm':
				*pMaxTids = atoi(optarg);
				break;
			case 'w':
				*pWorkMultiplier = atoi(optarg);
				break;
		}	
	}
	
	
}

/****************************************************************************
Function: 		main

Description: 	Launch a large number of threads, wait on each Thread and
							display the calculated sum as well as how many times each
							thread updated the global sum.
							
Parameters: 	argc, argv: take command line options:
							-x turn off synchronization
							-m VALUE : set number of threads to VALUE
							-w VALUE : set work multiplier to VALUE

Returned: 		int - 0 if success, -1 if failure
****************************************************************************/
int main(int argc, char** argv)
{
	int i = 0;
	int numbThreads;

	pthread_attr_t sAttr;
	pthread_t *paTids;
	int *paThreadParams;
	
	// default values
	gbSynchronizationOn = true;
	gMaxTids = 1;
	gWorkMultiplier = 100;
	
	parseCommandLineOptions(argc, argv, &gbSynchronizationOn,
		&gMaxTids, &gWorkMultiplier);
	
	numbThreads = gMaxTids;
	
	paTids = (pthread_t*) malloc(sizeof(pthread_t) * numbThreads);
	paThreadParams = (int*) malloc(sizeof(int) * numbThreads);


	gpaThreadExecutionCounter = (int*) malloc(sizeof(int) * numbThreads);
	
	pthread_attr_init(&sAttr);

	
	// allocate data to send to the threads
	for(i=0; i<numbThreads; i++)
	{
		paThreadParams[i] = i;
		gpaThreadExecutionCounter[paThreadParams[i]] = 0;
	}

	// create threads
	for(i=0; i<numbThreads; i++)
	{
		pthread_create(&(paTids[i]), &sAttr, runner, &(paThreadParams[i]));
	}
	
	
	// wait on each thread
	for(i=0; i<numbThreads; i++)
	{
		pthread_join(paTids[i], NULL);
	}

	// the global solution
	printf("gSum: %Lf\n\n", gSum);


	// display how much work each thread performed
	for(i=0; i < numbThreads; i++)
	{
		printf("[%3d]:%3d\t", i,gpaThreadExecutionCounter[i]);
		if( (i+1) % 10 == 0)
		{
			printf("\n");
		}
	}

	
	printf("\n\n");

	pthread_attr_destroy(&sAttr);

	free(gpaThreadExecutionCounter);
	free(paThreadParams);
	free(paTids);
	
	return EXIT_SUCCESS;
}
